import yaml


class LatestUpdate:

    def __init__(self, config):
        self.requirements = config.pop("requirements")
        self.dependencies = config.pop("dependencies")

    @classmethod
    def create(cls, requirements, dependencies):
        return cls({
            "requirements": requirements,
            "dependencies": dependencies
        })

    @classmethod
    def read(cls, path):
        with open(path, "r") as f:
            yml = yaml.load(f, yaml.SafeLoader) or {}
        return cls(yml)

    def write(self, path):
        with open(path, "w") as f:
            yaml.dump(vars(self), f)



import logging
import os
from pathlib import Path
from typing import List, Dict

import yaml
from travel.config.bag import Bag

BAG_FILE = "bag.yml"


logger = logging.getLogger(__name__)


def _read_bag(location: str) -> Bag:

    # Read the bag file
    path = os.path.join(location, BAG_FILE)
    with open(path) as f:
        yml = yaml.load(f, Loader=yaml.SafeLoader) or {}
        return Bag(location=location, yml=yml)


def _has_bag(location: str) -> bool:
    return os.path.isfile(os.path.join(location, BAG_FILE))


def get_bag_name(location: str) -> str:
    return os.path.basename(os.path.normpath(location))


def read_all_bags(location: str) -> Dict[str, Bag]:

    # Find the uppermost (parent) bag file
    uppermost = location
    while _has_bag(location):
        uppermost = location
        location = str(Path(location).parent)

    # Read main bag file and baged bag files
    bags = {bag.name: bag for bag in _read_bags_from(uppermost)}
    # Set the root context
    for b in bags.values():
        b.root_context = uppermost

    # Build dependencies
    for name in bags:
        for bag in bags.values():
            if name in bag.dependencies:
                bag.fill_dependency_with_bag(bags[name])
    return bags


def parse_bags(location: str, target: str = None) -> (Bag, Bag):

    # Read the target bag and all the bags
    target = target or get_bag_name(location)
    bags = read_all_bags(location)

    # Manage this target bag
    if target not in bags:
        raise ValueError(f"The specified bag \"{target}\" does not exist in {location}")
    bag = bags[target]
    return bag, bags


def _read_bags_from(location: str) -> List[Bag]:

    # Read the local bag
    group = []
    bag = _read_bag(location)

    # Read all bags recursively
    bags = [bag]
    for directory in os.listdir(location):
        current = os.path.join(location, directory)
        if _has_bag(current):
            children = _read_bags_from(current)
            group = group + children
            bags = bags + children

    # Set the baged group
    bag.group = group
    return bags



import argparse
import logging
import os

from travel.cli import python_wrapper, packer, planner, adder
from travel.cli.cleaner import Cleaner
from travel.cli.setupper import Setupper

logger = logging.getLogger(__name__)


def main():

    # Travel
    parser = argparse.ArgumentParser()
    parser.add_argument("--debug", help="Print a stacktrace of errors", action="store_true")
    parser.add_argument("--context", default=os.getcwd(), help="Path to the folder containing the bag.yml file")
    parser.set_defaults(action=lambda args: parser.parse_args(["-h"]))
    subparsers = parser.add_subparsers()

    # Clean
    clean = subparsers.add_parser("clean")
    clean.set_defaults(action=lambda args, rest: Cleaner().manage(args.context))

    # Plan
    plan = subparsers.add_parser("plan")
    plan.add_argument("name", help="Main bag name (name of the project)")
    plan.set_defaults(action=lambda args, rest: planner.run(args.context, args.name))

    # Add
    plan = subparsers.add_parser("add")
    plan.add_argument("plan", help="Plan URL or path")
    plan.add_argument("--checkout", help="Name of the branch to checkout", required=False)
    plan.add_argument("--directory", help="Name of the directory in the plan to use", required=False)
    plan.add_argument("--no-input", help="Do not ask for inputs", required=False, action="store_true")
    plan.add_argument("--config", help="Config in key=value format", nargs="+", required=False)
    plan.set_defaults(action=lambda args, rest: adder.run(args.context, args.plan, checkout=args.checkout, directory=args.directory, no_input=args.no_input, config=args.config))

    # Setup
    setup = subparsers.add_parser("setup")
    setup.set_defaults(action=lambda args, rest: Setupper().manage(args.context))

    # Pack
    pack = subparsers.add_parser("pack")
    pack.add_argument("--target", help="Name of the bag to run setup.py commands", required=False)
    pack.add_argument("--no-setup", help="Do not update the venvs", required=False, action="store_true")
    pack.set_defaults(action=lambda args, rest: packer.pack(args.context, rest, target=args.target, setup=not args.no_setup))

    # Release

    # # Python
    # python = subparsers.add_parser("python")
    # python.add_argument("package", help="Name of the bag (it will be used to activate its venv)")
    # python.set_defaults(action=lambda args, rest: python_wrapper.run(args.context, args.package, rest))

    # Parse args
    arguments, remainder = parser.parse_known_args()

    try:

        # Invoke the corresponding functions
        arguments.action(arguments, remainder)
        logger.info("All done.")

    except Exception as e:

        if arguments.debug:
            logger.exception("")
        else:
            logger.error(e if str(e) else f"Error: {type(e)}")
        exit(1)


if __name__ == '__main__':
    main()



import os
import re
from typing import Optional

from travel.config.sanitizers import asserter


LATEST_PIP = "latest"

_BASE_CHARS = r"[A-Za-z0-9\-\.\[,\]_]+"  # Support the package[extra1,extra2] syntax
_PACKAGE_NAME = re.compile(_BASE_CHARS)
_PACKAGE = re.compile(r"^"+_BASE_CHARS+r"$")
_VERSION = re.compile(r"^"+_BASE_CHARS+r"$")  # To be changed or not?

_MODIFIERS = r"(==|<|<=|>|>=|!=|===|~=)"
_VERSIONS_M = r"[A-Za-z0-9\-\._\*]+"
_VERSIONED_PACKAGE = re.compile(r"^"+_BASE_CHARS+r"("+_MODIFIERS+_VERSIONS_M+", *)*("+_MODIFIERS+_VERSIONS_M+")? *$")


def sanitize_package(package: str) -> str:
    return asserter.regex(_PACKAGE, package)


def is_just_package(package: str) -> bool:
    return bool(_PACKAGE.match(package))


def sanitize_version(version: str, accept_path=False) -> str:
    if accept_path and os.path.isdir(version):
        return version
    else:
        return asserter.regex(_VERSION, version)


def sanitize_pip_version(version: str, nullable: bool = False) -> Optional[str]:
    if version is None and nullable:
        return None
    else:
        return version == LATEST_PIP or asserter.regex(_VERSION, version)


def sanitize_versioned_package(spec: str) -> str:
    return asserter.regex(_VERSIONED_PACKAGE, spec)


def get_package_name(spec: str, standardize: bool = False) -> str:
    return _PACKAGE_NAME.match(spec).group(0).lower().replace("-", "_")



import itertools
import json
import logging
import os
import shutil
from distutils.dir_util import copy_tree
from pathlib import Path

import setuptools
from travel.cli.setupper import Setupper
from travel.config.bag import Bag
from travel.config.reader import parse_bags
from travel.custom.tasks import performer
from travel.tools.python import main_python

logger = logging.getLogger(__name__)

MANIFEST = "MANIFEST.in"
SETUP_PY = "setup.py"


def _find_code_and_data(dep: Bag):

    # Get all the files that should be in the dependency (code and package_data)
    egg_info = f"{dep.package}.egg-info"
    sources = os.path.join(dep.setup_py_folder, egg_info, "SOURCES.txt")  # This file is created by setup
    with open(sources, "r") as f:
        files = f.read().splitlines()

    # Get just the unique folders starting with the package name (and are not egg info)
    folders = set([
        os.path.dirname(f) for f in files
        if f.startswith(dep.package) and not f.startswith(egg_info)
    ])

    # Find code packages
    code_packages = set([p for p in setuptools.find_packages(where=dep.setup_py_folder) if "." not in p])

    # Find package_data
    package_data = folders - code_packages

    # Return the list
    return list(code_packages), list(package_data)


def create_egg_info(current_bag: Bag):
    for bag in current_bag.flat_dependencies(with_current=True):
        setup_py = os.path.join(bag.setup_py_folder, SETUP_PY)
        main_python.run(f"{setup_py} egg_info", cwd=bag.setup_py_folder)


def pack(context: str, command: str, target: str = None, setup: bool = True):

    # Setup the bags and dependencies
    if setup:
        current_bag, all_bags = Setupper().manage(context, target=target)
    else:
        # If no setup, create just the egg_info folder for SOURCES.txt information for each dependency (and current)
        current_bag, all_bags = parse_bags(context, target)
        create_egg_info(current_bag)

    # Pre-pack
    performer.perform_tasks("pack", "pre", current_bag)

    # Clean the previous target folder, if existing
    build_folder = current_bag.build_folder
    if os.path.isdir(build_folder):
        shutil.rmtree(build_folder)

    # Copy the structure of this package
    _copy_folder(current_bag.setup_py_folder, build_folder)

    # Copy the MANIFEST.in
    source_build_folder = os.path.join(build_folder, os.path.basename(current_bag.setup_py_folder))
    manifest_file = os.path.join(source_build_folder, MANIFEST)
    Path(manifest_file).touch(exist_ok=True)

    # For all dependencies, copy their code and package_data too
    manifest_first_time = True
    for dep in current_bag.flat_dependencies():

        # Find code packages and package_data
        code, data = _find_code_and_data(dep)

        # Copy it inside the copied setup.py folder
        for folder in code:
            _copy_folder(
                os.path.join(dep.setup_py_folder, folder),
                os.path.join(source_build_folder)
            )

        # Store the package_data information
        manifest = [f"include {d}/*" for d in data]
        with open(manifest_file, "a") as f:
            # Write the comment for ease of documentation
            if manifest and manifest_first_time:
                f.write(f"\n#travel-dependency")
                manifest_first_time = False
            # Add each package_data
            f.write("\n"+"\n".join(manifest))  # Add a \n in case file is not ending with that

    # Setup the code
    setup_py = os.path.join(source_build_folder, SETUP_PY)
    main_python.run(f"{setup_py} {' '.join(command)}", cwd=source_build_folder)  # TODO should check for spaces and commas, or use list!

    # Post-pack
    performer.perform_tasks("pack", "post", current_bag)


def _copy_folder(source, destination):
    os.makedirs(destination, exist_ok=True)
    right_destination = os.path.join(destination, os.path.basename(os.path.normpath(source)))
    return copy_tree(source, right_destination)



import logging
import os
import shutil

from travel.cli.base import TravelCommand
from travel.config.bag import Bag
from travel.custom.scopes.scoped_venvs import ScopedVirtualenvs
from travel.tools.venv import Virtualenv

logger = logging.getLogger(__name__)


class Cleaner(TravelCommand):

    def _phase_name(self):
        return "clean"

    def _perform_tasks(self, bag: Bag, step: str):
        if step == "post":
            pass
        else:
            super()._perform_tasks(bag, step)

    def _manage(self, bag: Bag):

        logger.info(f"Cleaning {bag.name}")
        to_remove = [
            os.path.join(bag.setup_py_folder, f"{bag.package}.egg-info"),     # egg-info
            Virtualenv(bag).path,                                              # venv folder
            bag.build_folder                                                   # artifacts build folder
        ]

        # If with scopes, remove them too
        if bag.scopes:
            scopes = ScopedVirtualenvs(bag)
            for env in scopes.envs.values():
                to_remove.append(env.path)

        # Clean
        for path in to_remove:
            if os.path.isdir(path):
                shutil.rmtree(path)




import logging
import os
import shutil
from pathlib import Path
from typing import List, Dict, Any

import yaml
from cookiecutter.exceptions import OutputDirExistsException
from cookiecutter.main import cookiecutter
from travel.config.reader import BAG_FILE
from travel.config.sanitizers import name_sanitizer

logger = logging.getLogger(__name__)
_PLAN = "plan"
_TRAVEL_FILE = "travel.yml"
_CONFIG = "config"
_CHECKOUT = "checkout"
_DIRECTORY = "directory"


def _is_a_plan(properties):
    return _PLAN in properties and isinstance(properties[_PLAN], str)


def _generate_breath_first(folder: str, yml: Dict[str, Any]):

    # Generate these plans
    for bag, properties in yml.items():

        # Get the folder where the bag will be created
        name_sanitizer.sanitize_name(bag)

        # Is it a plan?
        if _is_a_plan(properties):

            # Create it
            plan = properties[_PLAN]
            cookiecutter(
                template=plan,
                output_dir=folder,
                no_input=True,
                extra_context=properties[_CONFIG],
                checkout=properties.get(_CHECKOUT),
                directory=properties.get(_DIRECTORY)
            )

    # Create the "folder" bag.yml file
    bag_file = os.path.join(folder, BAG_FILE)
    if not os.path.isfile(bag_file):
        Path(bag_file).touch()

    # Generate the subbags
    for bag, properties in yml.items():
        bag_folder = os.path.join(folder, bag)
        if not _is_a_plan(properties):
            _generate_breath_first(bag_folder, properties)


def run(context: str, name: str):

    # Read the travel file
    travel_file = os.path.join(context, _TRAVEL_FILE)
    with open(travel_file) as f:
        yml = yaml.load(f, Loader=yaml.SafeLoader)
    if not yml:
        raise ValueError("Travel file must not be empty.")

    # Generate
    folder = os.path.join(context, name)
    try:
        _generate_breath_first(folder, yml)
    except OutputDirExistsException as e:
        # If the error is just because the folder exists, exit
        raise e
    except Exception as e:
        # Remove in case of any other error
        if Path(folder).is_dir():
            shutil.rmtree(folder)
        raise e



import os
from pathlib import Path

from travel.cli.base import TravelCommand
from travel.config.bag import Bag
from travel.custom.scopes.scoped_venvs import ScopedVirtualenvs
from travel.tools.venv import Virtualenv


class Setupper(TravelCommand):

    def _phase_name(self) -> str:
        return "setup"

    def _manage(self, bag: Bag):

        # Create the virtualenv
        venv = Virtualenv(bag, touch_requirements_file=True)
        venv.create()
        updated = venv.update()

        # Pip freeze
        if updated:
            venv.freeze()

        # Create the scopes
        if bag.scopes:
            scopes = ScopedVirtualenvs(bag, touch_requirements_file=True)
            for scope in bag.scopes:
                scopes.create(scope)
                updated = scopes.update(scope)
                if updated:
                    scopes.freeze(scope)




import os

from travel.config.bag import Bag
from travel.tools.base_venv import BaseVirtualenv
from travel.tools.python import Python
from travel.tools.python import main_python as default_python


class ScopedVirtualenvs():

    def __init__(self, bag: Bag, main_python: Python = default_python, touch_requirements_file: bool = False):
        self.envs = {
            scope: BaseVirtualenv(
                location=bag.location,
                name_suffix=f"{bag.name}-{scope}",
                pip_config=bag.pip,
                dependencies=bag.flat_dependencies(with_current=True),
                requirements_file=os.path.join(bag.location, f"requirements_{scope}.txt"),
                main_python=main_python,
                touch_requirements_file=touch_requirements_file,
                extra_requirements=config.requirements
            )
            for scope, config in bag.scopes.items()
        }

    def create(self, scope: str):
        self.envs[scope].create()

    def create_all(self):
        for scope in self.envs.keys():
            self.create(scope)

    def update(self, scope: str) -> bool:
        return self.envs[scope].update()

    def update_all(self):
        for scope in self.envs.keys():
            self.update(scope)

    def freeze(self, scope: str):
        self.envs[scope].freeze()

    def freeze_all(self):
        for scope in self.envs.keys():
            self.freeze(scope)



import logging
from typing import List

from cookiecutter.main import cookiecutter

logger = logging.getLogger(__name__)


def _validate(config):
    if "=" not in config:
        raise ValueError(f"Configs should have '=', so this is not valid: {config}")
    return config


def run(context: str, plan: str, checkout: str = None, directory: str = None, no_input: bool = False, config: List[str] = None):

    extra_context = {
        key: value
        for key, value in [
            _validate(c).split("=", maxsplit=1)
            for c in (config if config else [])
        ]
    }

    cookiecutter(
        template=plan,
        output_dir=context,
        checkout=checkout,
        directory=directory,
        no_input=no_input,
        extra_context=extra_context
    )



import logging
import os
import re
import subprocess
from pathlib import Path
from typing import List

from travel.config.bag import Bag
from travel.config.sanitizers import pip_sanitizer
from travel.config.sanitizers.pip_sanitizer import LATEST_PIP
from travel.config.subconfigs.pip import PipConfig
from travel.tools.outputs.latest_updates import LatestUpdate
from travel.tools.pip import Pip
from travel.tools.python import Python
from travel.tools.python import main_python as default_python
from travel.tools.shell import Shell

logger = logging.getLogger(__name__)


_CREATE_VENV = "{python} -m venv {venv}"
LATEST_UPDATE = "latest_update.yml"


class BaseVirtualenv:

    def __init__(self, location: str, name_suffix: str, pip_config: PipConfig, dependencies: List[Bag], requirements_file: str, main_python: Python = default_python, touch_requirements_file: bool = False, extra_requirements: List[str] = None):
        self._main_python = main_python
        self._name = f"venv-{name_suffix}"

        self.pip_config = pip_config
        self.dependencies = dependencies
        self.requirements_file = requirements_file
        self.extra_requirements = extra_requirements

        self.location = location
        self.path = os.path.join(location, self._name)
        self.python = Python(path=os.path.join(self.path, "bin" if os.name == "posix" else "Scripts", "python" if os.name == "posix" else "python.exe"))
        self.pip = Pip(self.python)

        # Prepare the requirements file
        if touch_requirements_file:
            if not os.path.isfile(requirements_file):
                Path(requirements_file).touch()

    def create(self) -> None:
        # Create the virtualenv if it does not exist
        if os.path.exists(self.path):
            logger.info(f"Virtualenv already exists for {self._name}")
        else:
            # Create the virtualenv with the template command
            logger.info(f"Creating the virtual environment for {self._name}...")
            Shell().run(
                _CREATE_VENV.format(
                    python=self._main_python.path,
                    venv=self.path
                )
            )

    def update(self) -> bool:

        # Upgrade pip if required
        pip_version = self.pip_config.version
        if pip_version:
            logger.info("Check pip version...")
            if pip_version == LATEST_PIP:
                self.pip.run("install --upgrade pip")
            else:
                self.pip.install(f"pip=={pip_version}")

        # Compute current requirements
        dependency_names = [bag.name for bag in self.dependencies]
        explicit_requirements = [req for bag in self.dependencies for req in bag.requirements]
        if self.extra_requirements:
            explicit_requirements = explicit_requirements + self.extra_requirements

        # If there is already a requirements file and a latest_update file, check: we might skip the update!
        latest_update_path = Path(self.path)/LATEST_UPDATE
        if Path(self.requirements_file).is_file() and latest_update_path.is_file():

            # Get the freezed requirements and the written requirements
            freezed_requirements = self._freeze()
            with open(self.requirements_file, "r") as f:
                written_requirements = f.read().splitlines()
            
            # Read the latest_update
            latest_update = LatestUpdate.read(latest_update_path)

            # If there is nothing to update, exit!
            if (set(freezed_requirements) == set(written_requirements)
                    and set(dependency_names) == set(latest_update.dependencies)
                    and set(explicit_requirements) == set(latest_update.requirements)):
                logger.info("Virtualenv is already up-to-date!")
                return False

        # Install requirements
        logger.info("Installing requirements...")
        for bag in self.dependencies:
            # Install the bag's package
            self.pip.run(f"install -e {bag.setup_py_folder}")
            # Install the explicit requirements
            if bag.requirements:
                self.pip.install(bag.requirements)

        if self.extra_requirements:
            self.pip.install(self.extra_requirements)

        # Uninstall the unnecessary requirements
        installed_requirements = [pip_sanitizer.get_package_name(f) for f in self._freeze()]
        necessary_requirements = self._recursive_requirements(explicit_requirements) if explicit_requirements else []
        unnecessary_requirements = set(installed_requirements) - set(necessary_requirements)
        if unnecessary_requirements:
            to_uninstall = ' '.join(list(unnecessary_requirements))
            logger.warning(f"Unnecessary requirements found, they will be uninstalled: {to_uninstall}")
            self.pip.run(f"uninstall -y {to_uninstall}")

        # Save the latest update
        LatestUpdate.create(
            requirements=explicit_requirements,
            dependencies=dependency_names
        ).write(latest_update_path)

        logger.info("Done.")
        return True

    def _freeze(self) -> List[str]:

        try:

            # Pip freeze
            output = self.pip.run(f"freeze", capture=True).stdout
            requirements = re.sub("[\r\n]+", "\n", output).split("\n")

            # Get real requirements
            return [req for req in requirements if _is_valid_requirement(req)]

        except subprocess.CalledProcessError as e:
            logger.error(e.stderr)
            exit(e.returncode)

    def freeze(self):

        # Update requirements file
        real_requirements = self._freeze()
        with open(self.requirements_file, "w") as f:
            for requirement in real_requirements:
                f.write(requirement + "\n")
            logger.info(f"Freezing {real_requirements}")

    def _recursive_requirements(self, requirements: List[str]) -> List[str]:

        def _find_recursively(reqs):

            # Parse this level requirements
            explicit_requirements = [pip_sanitizer.get_package_name(req) for req in reqs]

            # Get the implicit requirements, needed by these requirements
            output = self.pip.run(f"show {' '.join(explicit_requirements)}", capture=True).stdout
            implicit = [
                req
                for line in re.sub("[\r\n]+", "\n", output).split("\n") if line.startswith("Requires: ")
                for req in line.replace("Requires: ", "").split(", ") if req
            ]

            # If any implicit, recursively get their implicit ones; else return it simply
            if implicit:
                requires = explicit_requirements + _find_recursively(implicit)
            else:
                requires = explicit_requirements
            return requires

        return list(set(_find_recursively(requirements)))


def _is_valid_requirement(requirement: str) -> bool:

    # List of requirements to filter out
    forbidden_requirements = [
        "pkg-resources",
        "^-e",
        "^#",
        "pywin32",
        "pywinpty"
    ]
    forbidden_match = "(" + ")|(".join(forbidden_requirements) + ")"

    # Check if there is one match
    return requirement and not re.match(forbidden_match, requirement)



from typing import List, Union

from travel.config.reader import read_all_bags
from travel.config.sanitizers.pip_sanitizer import sanitize_package, sanitize_versioned_package, is_just_package
from travel.tools.executable import Executable
from travel.tools.python import Python


class Pip(Executable):

    def __init__(self, python: Python, pre_command: str = None):
        super().__init__(f"{python.path} -m pip --disable-pip-version-check", pre_command=pre_command)

    def install(self, requirements: Union[List[str], str], allow_bags_from: Union[str, List[str]] = None):

        if isinstance(requirements, str):
            requirements = [requirements]

        # Install real requirements
        packages = [
            f'{sanitize_versioned_package(req)}'
            for req in requirements
            if not is_just_package(req)
        ]
        if packages:
            super().run(f"install {' '.join(packages)}")

        # Install bags
        bags = [
            f'{sanitize_package(req)}'
            for req in requirements
            if is_just_package(req)
        ]
        if bags:
            if not allow_bags_from:
                raise ValueError(f"{bags} without versions (or bags) are not allowed here")
            else:
                allow_bags_from = allow_bags_from if isinstance(allow_bags_from, list) else [allow_bags_from]
                for context in allow_bags_from:
                    all_bags = read_all_bags(context)
                    super().run(f"install -e {' '.join([b.setup_py_folder for b in all_bags.values() if b.name in bags])}")



import logging

from travel.config.sanitizers import pip_sanitizer

logger = logging.getLogger(__name__)


class PipConfig:

    def __init__(self, config: dict):
        self.version = pip_sanitizer.sanitize_pip_version(config.pop("version", None), nullable=True)

        # If there are still configs, they are unknown. Print a warning
        if config:
            logger.warning(f"Unknown configuration for pip: {config}")



import abc
import logging
from typing import Set

from travel.config.bag import Bag
from travel.config.reader import get_bag_name, read_all_bags, parse_bags
from travel.custom.tasks import performer
from travel.tools.python import main_python, Python

logger = logging.getLogger(__name__)


class TravelCommand(abc.ABC):

    def __init__(self, python: Python = main_python):
        self.main_python = python

    @abc.abstractmethod
    def _phase_name(self) -> str:
        pass

    def _perform_tasks(self, bag: Bag, step: str):
        return performer.perform_tasks(self._phase_name(), step, bag)

    @abc.abstractmethod
    def _manage(self, bag: Bag):
        pass

    def _manage_from_bag_recursive(self, bag: Bag, done: Set[Bag]):

        if not bag.group:

            # Manage dependencies first, then the bag
            for b in [*bag.flat_dependencies(), bag]:
                if b not in done:
                    logger.info("")
                    logger.info("="*60)
                    logger.info(f"=== {b.name.center(52, ' ')} ===")
                    logger.info("="*60)
                    self._perform_tasks(b, "pre")
                    self._manage(b)
                    self._perform_tasks(b, "post")
                    logger.info("=" * 60)
                    logger.info("")
                    logger.info("")
                    done.add(b)

        else:

            # Manage each element of the group
            for bag in bag.group:
                self._manage_from_bag_recursive(bag, done=done)

    def manage_from_bag(self, bag: Bag):
        self._manage_from_bag_recursive(bag, done=set())

    def manage(self, context: str, target: str = None) -> (Bag, Bag):
        bag, bags = parse_bags(context, target)
        self.manage_from_bag(bag)
        return bag, bags



import logging

from travel.config.sanitizers import pip_sanitizer


logger = logging.getLogger(__name__)


class ScopeConfig:

    def __init__(self, name: str, config: dict):
        self.name = name
        self.requirements = [pip_sanitizer.sanitize_versioned_package(req) for req in config.pop("requirements", {})]

        # If there are still configs, they are unknown. Print a warning
        if config:
            logger.warning(f"Unknown configuration for scope \"{name}\": {config}")



from travel.config.bag import Bag
from travel.tools.base_venv import BaseVirtualenv
from travel.tools.python import Python
from travel.tools.python import main_python as default_python


class Virtualenv(BaseVirtualenv):

    def __init__(self, bag: Bag, main_python: Python = default_python, touch_requirements_file: bool = False):
        super().__init__(
            location=bag.location,
            name_suffix=bag.name,
            pip_config=bag.pip,
            dependencies=bag.flat_dependencies(with_current=True),
            requirements_file=bag.requirements_file,
            main_python=main_python,
            touch_requirements_file=touch_requirements_file
        )
